#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd


from kxy.api import mutual_information_analysis
from kxy.api.core import least_continuous_mutual_information, least_mixed_conditional_mutual_information, \
	scalar_continuous_entropy, spearman_corr, pearson_corr


def regression_achievable_performance_analysis(x_c, y, x_d=None, space='dual'):
	"""
	.. _regression-achievable-performance-analysis:
	Quantifies the :math:`R^2` that can be achieved when trying to predict :math:`y` with :math:`x`.

	.. math::

		\\bar{R}^2 = 1 - e^{-2I(y; f(x))}


	Parameters
	----------
	x_c : (n,d) np.array
		Continuous inputs.
	y : (n,) np.array
		Labels.
	x_d : (n, d) np.array or None (default), optional
		Discrete inputs.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.


	Returns
	-------
	a : pandas.DataFrame

		Dataframe with columns:

		* :code:`'Achievable R^2'`: The highest :math:`R^2` that can be achieved by a regression model using provided inputs.
		* :code:`'Achievable Log-Likelihood Per Sample'`: The highest true log-likelihood per sample that can be achieved by a regression model using provided inputs.
		

	.. admonition:: Theoretical Foundation

		Section :ref:`1 - Achievable Performance`.
	"""
	assert len(y.shape) == 1 or y.shape[1] == 1, 'y should be a one dimensional numpy array'

	if x_d is None:
		d = x_c.shape[1] if len(x_c.shape) > 1 else 1
		if d == 1:
			data = np.hstack((y[:, None], x_c[:, None], np.abs(x_c[:, None]-np.nanmean(x_c[:, None], axis=0))))
		else:
			data = np.hstack((y[:, None], x_c, np.abs(x_c-np.nanmean(x_c, axis=0))))	
		corr = pearson_corr(data) if space == 'primal' else spearman_corr(data)
		mi_analysis = mutual_information_analysis(corr, 0, space=space, greedy=True)
		mi = mi_analysis['mutual_information']

	else:
		y_ = y[:, None] if len(y.shape) == 1 else y
		x_d_ = np.array(['*_*'.join(list(r)) for r in x_d]) if len(y.shape) > 1 else x_d

		# I(y; x_d)
		mi = least_mixed_conditional_mutual_information(y, x_d_, space=space, non_monotonic_extension=False)
		categories = list(set(list(x_d_)))
		n = x_d_.shape[0]
		probas = np.array([1.*len(x_d_[x_d_==cat])/n for cat in categories])

		# I(y; x_c|x_d)
		mi += np.sum([probas[i]*least_continuous_mutual_information(\
			x_c[x_d_==categories[i]], y[x_d_==categories[i]], space=space) for i in range(len(categories))\
			if probas[i] > 0.0])

	hy = scalar_continuous_entropy(y, space=space)

	return pd.DataFrame({\
		'Achievable R^2': [1.-np.exp(-2.*mi)], \
		'Achievable True Log-Likelihood Per Sample': [hy-mi]})



def regression_variable_selection_analysis(x_c, y, x_d=None, space='dual'):
	"""
	.. _regression-variable-selection-analysis:
	Runs the variable selection analysis for a regression problem.


	Parameters
	----------
	x_c : (n,d) np.array
		Continuous inputs.
	y : (n,) np.array
		Labels.
	x_d : (n, d) np.array or None (default), optional
		Discrete inputs.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.


	Returns
	-------
	a : pandas.DataFrame

		Dataframe with columns:

		* :code:`'Variable'`: The variable index starting from 0 at the leftmost column of :code:`x_c` and ending at the rightmost column of :code:`x_d`.
		* :code:`'Selection Order'`: The order in which the associated variable was selected, starting at 1 for the most important variable.
		* :code:`'Univariate Achievable R^2'`: The highest :math:`R^2` that can be achieved by a regression model solely using this variable.
		* :code:`'Maximum Marginal R^2 Increase'`: The highest amount by which the :math:`R^2` can be increased as a result of adding this variable in the variable selection scheme.
		* :code:`'Running Achievable R^2'`: The highest :math:`R^2` that can be achieved by a regression model using all variables selected so far, including this one.
		* :code:`'Conditional Mutual Information (nats)'`: The mutual information between this variable and the label, conditional on all variables previously selected.
		* :code:`'Running Mutual Information (nats)'`: The mutual information between all variables selected so far, including this one, and the label.
		* :code:`'Maximum Marginal True Log-Likelihood Increase Per Sample'`: The highest amount by which the true log-likelihood per sample can increase as a result of adding this variable.
		* :code:`'Running Maximum Log-Likelihood Increase Per Sample'`: The highest amount by which the true log-likelihood per sample can increase (over the log-likelihood of the naive strategy consisting of predicting the mode of :math:`y`) as a result of using all variables selected so far, including this one.


	.. admonition:: Theoretical Foundation

		Section :ref:`2 - Variable Selection Analysis`.
	"""
	assert x_d is None, 'Variable selection for regression does not yet support discrete variables'

	d = x_c.shape[1] if len(x_c.shape) > 1 else 1
	data = np.hstack((y[:, None] , x_c, np.abs(x_c-np.nanmean(x_c, axis=0))))
	corr = pearson_corr(data) if space == 'primal' else spearman_corr(data)
	mi_analysis = mutual_information_analysis(corr, 0, space=space, greedy=True)
	columns = [0] + [_ for _ in range(d)] + [_ for _ in range(d)]

	if mi_analysis is None:
		return None

	remaining_columns = [_ for _ in range(d)]

	cmis = {}
	rsqs = {}
	mis = {}
	order = {}
	idx = 1
	for i in range(1, 1+2*d):
		column_id = mi_analysis['selection_order'][str(i)]
		column = columns[column_id]
		if column in remaining_columns:
			mis[column] = mi_analysis['individual_mutual_informations'][str(i)]
			if idx == 1:
				cmis[column] = mis[column]
			else:
				cmis[column] = mi_analysis['conditional_mutual_informations'][str(i)]
			rsqs[column] = 1.-np.exp(-2.*mis[column])
			order[column] = idx
			idx += 1
			remaining_columns.remove(column)

	rsq_inc = {}
	run_mi = 0
	run_rsqs = {}
	for v, o in sorted(order.items(), key=lambda item: item[1]):
		run_mi += cmis[v]
		run_rsqs[o] = 1.-np.exp(-2.*run_mi)

		if rsq_inc == {}:
			rsq_inc[o] = 1.-np.exp(-2.*cmis[v])
			
		else:
			rsq_inc[o] = run_rsqs[o]-run_rsqs[o-1]

	n = y.shape[0]
	max_r_2 = np.max([_ for _ in run_rsqs.values()])
	importance_df = pd.DataFrame({
		'Variable': [v for v, o in sorted(order.items(), key=lambda item: item[1])], \
		'Selection Order': [o for v, o in sorted(order.items(), key=lambda item: item[1])], \

		# Achievable R^2 analysis
		'Univariate Achievable R^2': [rsqs[v] for v, o in sorted(order.items(), key=lambda item: item[1])], \
		'Maximum Marginal R^2 Increase': [rsq_inc[o] for v, o in sorted(order.items(), key=lambda item: item[1])], \
		'Running Achievable R^2': [run_rsqs[o] for v, o in sorted(order.items(), key=lambda item: item[1])], \
		'Running Achievable R^2 (%)': [100.*run_rsqs[o]/max_r_2 for v, o in sorted(order.items(), key=lambda item: item[1])], \

		# Conditional mutual information analysis
		'Univariate Mutual Information (nats)': [mis[v] for v, o in sorted(order.items(), key=lambda item: item[1])], \
		'Conditional Mutual Information (nats)': [cmis[v] for v, o in sorted(order.items(), key=lambda item: item[1])], \

		# The largest likelihood by which the likelihood can be increased as a result of adding this variable
		'Univariate Maximum True Log-Likelihood Increase Per Sample': [mis[v] for v, o in sorted(order.items(), key=lambda item: item[1])], \
		'Maximum Marginal True Log-Likelihood Increase Per Sample': [cmis[v] for v, o in sorted(order.items(), key=lambda item: item[1])]})

	importance_df['Running Mutual Information (nats)'] = importance_df['Conditional Mutual Information (nats)'].cumsum()
	max_mi = importance_df['Running Mutual Information (nats)'].max()
	importance_df['Running Mutual Information (%)'] = 100.*importance_df['Running Mutual Information (nats)']/max_mi

	# The largest value by which the lof likelihood can be increased by using the variables selected so far.
	importance_df['Running Maximum Log-Likelihood Increase Per Sample'] = importance_df['Maximum Marginal True Log-Likelihood Increase Per Sample'].cumsum()

	return importance_df





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
We define a custom :code:`kxy` `pandas accessor <https://pandas.pydata.org/pandas-docs/stable/development/extending.html>`_ below, 
namely the class :code:`KXYAccessor`, that extends the pandas DataFrame class with all our analyses, thereby allowing data scientists to tap into 
the power of the :code:`kxy` toolkit within the comfort of their favorite data structure.

All methods defined in the :code:`KXYAccessor` class are accessible from any DataFrame instance as :code:`df.kxy.<method_name>`, so long as the :code:`kxy` python 
package is imported alongside :code:`pandas`. 
"""

from functools import lru_cache, wraps
import logging
import os
from concurrent.futures import ThreadPoolExecutor

import numpy as np
import pandas as pd
import seaborn as sns

from kxy.asset_management import information_adjusted_beta, information_adjusted_correlation, \
	robust_pearson_corr

from kxy.api.core import spearman_corr, pearson_corr, auto_predictability

from kxy.classification import classification_achievable_performance_analysis, \
	classification_variable_selection_analysis, classification_model_improvability_analysis, \
	classification_model_explanation_analysis, classification_bias

from kxy.regression import regression_achievable_performance_analysis, regression_variable_selection_analysis, \
	regression_model_improvability_analysis, regression_model_explanation_analysis, regression_bias



@pd.api.extensions.register_dataframe_accessor("kxy")
class KXYAccessor(object):
	"""
	Extension of the pandas.DataFrame class with various analytics for **pre-learning** and **post-learning**,
	in supervised learning problems.
	"""
	def __init__(self, pandas_obj):
		self._obj = pandas_obj


	def is_discrete(self, column):
		"""
		Determine whether the input column contains discrete observations.
		"""
		ret = (not np.can_cast(self._obj[column].values, float))
		ret = ret or len(list(set(self._obj[column].values))) < 0.5*self._obj.shape[0]

		return ret


	def is_categorical(self, column):
		"""
		Determine whether the input column contains categorical observations.
		"""
		ret = (not np.can_cast(self._obj[column].values, float))

		return ret


	def corr(self, columns=(), method='information-adjusted', min_periods=1, p=0, p_ic='hqic'):
		"""
		Calculates the auto-correlation matrix of all columns or the input subset.


		Parameters
		----------
		columns : set, optional
			The set of columns to use. If not provided, all columns are used.
		method : str, optional
			Which method to use to calculate the auto-correlation matrix. Supported
			values are 'information-adjusted' (the default) and all 'method' values of pandas.DataFrame.corr.
		p : int, optional
			The number of auto-correlation lags to use as empirical evidence in the maximum-entropy problem. 
			The default value is 0, which corresponds to assuming rows are i.i.d. Values other than 0 are only
			supported in the robust-pearson method. When p is None, it is inferred from the sample.
		min_periods : int, optional
			Only used when method is not 'information-adjusted'. 
			See the documentation of pandas.DataFrame.corr.
		p_ic : str
			The criterion used to learn the optimal value of :code:`p` (by fitting a VAR(p) model) when :code:`p=None`. Should be one of 'hqic' (Hannan-Quinn Information Criterion), 'aic' (Akaike Information Criterion), 'bic' (Bayes Information Criterion) and 't-stat' (based on last lag). Same as the 'ic' parameter of :code:`statsmodels.tsa.api.VAR`.



		Returns
		-------
		c : pandas.DataFrame
			The auto-correlation matrix.


		.. seealso::

			:ref:`kxy.finance.risk_analysis.information_adjusted_correlation <information-adjusted-correlation>`
			:ref:`kxy.finance.risk_analysis.robust_pearson_corr <robust-pearson-corr>`
		"""
		columns = self._obj.columns if columns == () else list(columns)

		if method == 'information-adjusted':
			c = information_adjusted_correlation(self._obj[columns].values, y=None)
			return pd.DataFrame(c, columns=columns, index=columns)

		if method == 'robust-pearson':
			c = robust_pearson_corr(self._obj[columns].values, y=None, p=p, p_ic=p_ic)
			return pd.DataFrame(c, columns=columns, index=columns)

		if method == 'spearman':
			c = spearman_corr(self._obj[columns].values)
			return pd.DataFrame(c, columns=columns, index=columns)

		if method == 'pearson':
			c = pearson_corr(self._obj[columns].values)
			return pd.DataFrame(c, columns=columns, index=columns)

		else:
			return pd.DataFrame.corr(self._obj[columns], method=method, min_periods=min_periods)



	def beta(self, market_returns_column, asset_returns_columns=(), risk_free_column=None,\
			method='information-adjusted', p=0, p_ic='hqic'):
		"""
		Calculates the beta of a portfolio/asset (whose returns are provided in column_y) 
		with respect to the market (whose returns are provided in market_returns_column) using a variety
		of estimation methods including the standard OLS/Pearson methods and information theoretical 
		alternatives aiming at accounting for nonlinearities and memory in asset returns.


		Parameters
		----------
		asset_returns_columns : str or list of str
			The name(s) of the column(s) to use for portfolio/asset returns.
		market_returns_column : str
			The name of the column to use for market returns.
		method : str, optional
			One of 'information-adjusted', 'robust-pearson', 'spearman',  or 'pearson'. This is the method to use
			to estimate the correlation between portfolio/asset returns and market returns.
		p : int, optional
			The number of auto-correlation lags to use as empirical evidence in the maximum-entropy problem. 
			The default value is 0, which corresponds to assuming rows are i.i.d. Values other than 0 are only
			supported in the robust-pearson method. When p is None, it is inferred from the sample.
		p_ic : str
			The criterion used to learn the optimal value of :code:`p` (by fitting a VAR(p) model) when :code:`p=None`.
			Should be one of 'hqic' (Hannan-Quinn Information Criterion), 'aic' (Akaike Information Criterion),
			'bic' (Bayes Information Criterion) and 't-stat' (based on last lag). Same as the 'ic' parameter of 
			:code:`statsmodels.tsa.api.VAR`.


		Returns
		-------
		c : pandas.DataFrame
			The beta coefficient(s).


		.. seealso::

			:ref:`kxy.finance.factor_analysis.information_adjusted_beta <information-adjusted-beta>`
		"""
		asset_returns_columns = [_ for _ in self._obj.columns if _ != market_returns_column] if asset_returns_columns == () \
			else [asset_returns_columns] if type(asset_returns_columns) == str else list(asset_returns_columns)
		columns = [market_returns_column] + asset_returns_columns

		c = self.corr(method=method, columns=columns, p=p, p_ic=p_ic).values[0, 1:]
		betas = c * np.sqrt(np.nanvar(self._obj[asset_returns_columns].values, axis=0)/\
			np.nanvar(self._obj[market_returns_column].values))

		if type(asset_returns_columns) == str:
			res = pd.DataFrame({asset_returns_columns: betas}).T.rename(columns={0: 'beta'})
		else:
			res = pd.DataFrame({asset_returns_columns[i]: [betas[i]] \
				for i in range(len(asset_returns_columns))}).T.rename(columns={0: 'beta'})

		return res



	def achievable_performance_analysis(self, label_column, input_columns=(), space='dual'):
		"""
		Runs the achievable performance analysis on a trained supervised learning model.

		The nature of the supervised learning problem (i.e. regression or classification) is inferred from whether or not :code:`label_column` is categorical.


		Parameters
		----------
		label_column : str
			The name of the column containing true labels.
		input_columns : set
			List of columns to use as inputs. When an empty set/list is provided, all columns but :code:`label_column` are used as inputs.
		space : str, 'primal' | 'dual'
			The space in which the maximum entropy problem is solved. 
			When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
			When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.



		Returns
		-------
		res : pandas.Styler
			res.data is a pandas.Dataframe with columns (where applicable):

			* :code:`'Achievable R^2'`: The highest :math:`R^2` that can be achieved by a model using provided inputs to predict the label.
			* :code:`'Achievable Log-Likelihood Per Sample'`: The highest true log-likelihood per sample that can be achieved by a model using provided inputs to predict the label.
			* :code:`'Achievable Accuracy'`: The highest classification accuracy that can be achieved by a model using provided inputs to predict the label.



		.. admonition:: Theoretical Foundation

			Section :ref:`1 - Achievable Performance`.

		.. seealso::

			* :ref:`kxy.regression.regression_achievable_performance_analysis <regression-achievable-performance-analysis>`
			* :ref:`kxy.classification.classification_achievable_performance_analysis <classification-achievable-performance-analysis>`
		"""
		problem = 'classification' if self.is_discrete(label_column) else 'regression'
		columns = [col for col in self._obj.columns if col != label_column] if len(input_columns) == 0\
			else input_columns
		discrete_columns = [col for col in columns if self.is_categorical(col)]
		continuous_columns =  [col for col in columns if not self.is_categorical(col)]

		y = self._obj[label_column].values
		x_c = self._obj[continuous_columns].values if len(continuous_columns) > 0 else None
		x_d = self._obj[discrete_columns].values if len(discrete_columns) > 0 else None

		res = regression_achievable_performance_analysis(x_c, y, x_d=x_d, space=space) if problem == 'regression' \
			else classification_achievable_performance_analysis(x_c, y, x_d=x_d, space=space)
		res = res.style.hide_index()

		return res


	def variable_selection_analysis(self, label_column, input_columns=(), space='dual'):
		"""
		Runs the model variable selection analysis on a trained supervised learning model.

		The nature of the supervised learning problem (i.e. regression or classification) is inferred from whether or not :code:`label_column` is categorical.


		Parameters
		----------
		label_column : str
			The name of the column containing true labels.
		input_columns : set
			List of columns to use as inputs. When an empty set/list is provided, all columns but :code:`label_column` are used as inputs.
		space : str, 'primal' | 'dual'
			The space in which the maximum entropy problem is solved. 
			When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
			When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.


		Returns
		-------
		res : pandas.Styler
			res.data is a pandas.Dataframe with columns (where applicable):

				* :code:`'Variable'`: The column name corresponding to the input variable.
				* :code:`'Selection Order'`: The order in which the associated variable was selected, starting at 1 for the most important variable.
				* :code:`'Univariate Achievable R^2'`: The highest :math:`R^2` that can be achieved by a classification model solely using this variable.
				* :code:`'Maximum Marginal R^2 Increase'`: The highest amount by which the :math:`R^2` can be increased as a result of adding this variable in the variable selection scheme.
				* :code:`'Running Achievable R^2'`: The highest :math:`R^2` that can be achieved by a classification model using all variables selected so far, including this one.
				* :code:`'Univariate Achievable Accuracy'`: The highest classification accuracy that can be achieved by a classification model solely using this variable.
				* :code:`'Maximum Marginal Accuracy Increase'`: The highest amount by which the classification accuracy can be increased as a result of adding this variable in the variable selection scheme.
				* :code:`'Running Achievable Accuracy'`: The highest classification accuracy that can be achieved by a classification model using all variables selected so far, including this one.
				* :code:`'Conditional Mutual Information (nats)'`: The mutual information between this variable and the label, conditional on all variables previously selected.
				* :code:`'Running Mutual Information (nats)'`: The mutual information between all variables selected so far, including this one, and the label.
				* :code:`'Univariate Achievable True Log-Likelihood Per Sample'`: The highest true log-likelihood per sample that can be achieved by a classification model solely using this variable.
				* :code:`'Maximum Marginal True Log-Likelihood Per Sample Increase'`: The highest amount by which the true log-likelihood per sample can be increased as a result of adding this variable in the variable selection scheme.
				* :code:`'Running Achievable True Log-Likelihood Per Sample'`: The highest true log-likelihood per sample that can be achieved by a classification model using all variables selected so far, including this one.
				* :code:`'Maximum Marginal True Log-Likelihood Increase Per Sample'`: The highest amount by which the true log-likelihood per sample can increase as a result of adding this variable.
				* :code:`'Running Maximum Log-Likelihood Increase Per Sample'`: The highest amount by which the true log-likelihood per sample can increase (over the log-likelihood of the naive strategy consisting of predicting the mode of :math:`y`) as a result of using all variables selected so far, including this one.


		.. admonition:: Theoretical Foundation

			Section :ref:`2 - Variable Selection Analysis`.


		.. seealso::

			* :ref:`kxy.regression.regression_variable_selection_analysis <regression-variable-selection-analysis>`
			* :ref:`kxy.classification.classification_variable_selection_analysis <classification-variable-selection-analysis>`
		"""
		problem = 'classification' if self.is_discrete(label_column) else 'regression'
		columns = [col for col in self._obj.columns if col != label_column] if len(input_columns) == 0\
			else input_columns
		discrete_columns = [col for col in columns if self.is_categorical(col)]
		continuous_columns =  [col for col in columns if not self.is_categorical(col)]

		y = self._obj[label_column].values
		x_c = self._obj[continuous_columns].values.astype(float) if len(continuous_columns) > 0 else None
		x_d = self._obj[discrete_columns].values.astype(str) if len(discrete_columns) > 0 else None

		res = regression_variable_selection_analysis(x_c, y, x_d=x_d, space=space) if problem == 'regression' \
			else classification_variable_selection_analysis(x_c, y, x_d=x_d, space=space)

		variable_columns = continuous_columns + discrete_columns
		res['Variable'] = res['Variable'].map({i: variable_columns[i] for i in range(len(variable_columns))})
		res.set_index(['Variable'], inplace=True)

		cm = sns.light_palette("green", as_cmap=True)
		res = res.style.background_gradient(cmap=cm)

		return res


	def model_improvability_analysis(self, label_column, model_prediction_column, input_columns=(), 
			space='dual'):
		"""
		Runs the model improvability analysis on a trained supervised learning model.

		The nature of the supervised learning problem (i.e. regression or classification) is inferred from whether or not :code:`label_column` is categorical.


		Parameters
		----------
		label_column : str
			The name of the column containing true labels.
		model_prediction_column : str
			The name of the column containing labels predicted by the model.
		input_columns : set
			List of columns to use as inputs. When an empty set/list is provided, all columns but :code:`model_prediction_column` and :code:`label_column` are used as inputs.
		space : str, 'primal' | 'dual'
			The space in which the maximum entropy problem is solved. 
			When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
			When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.



		Returns
		-------
		res : pandas.Styler
			res.data is a pandas.Dataframe with columns (where applicable):

				* :code:`'Leftover R^2'`: The amount by which the trained model's :math:`R^2` can still be increased without resorting to additional inputs, simply through better modeling.
				* :code:`'Leftover Log-Likelihood Per Sample'`: The amount by which the trained model's true log-likelihood per sample can still be increased without resorting to additional inputs, simply through better modeling.
				* :code:`'Leftover Accuracy'`: The amount by which the trained model's classification accuracy can still be increased without resorting to additional inputs, simply through better modeling.


		.. admonition:: Theoretical Foundation

			Section :ref:`3 - Model Improvability`.

		.. seealso::

			* :ref:`kxy.regression.regression_model_improvability_analysis <regression-model-improvability-analysis>`
			* :ref:`kxy.classification.classification_model_improvability_analysis <classification-model-improvability-analysis>`
		"""
		problem = 'classification' if self.is_discrete(label_column) else 'regression'
		columns = [col for col in self._obj.columns if col != label_column and col != model_prediction_column] \
			if len(input_columns) == 0 else input_columns
		discrete_columns = [col for col in columns if self.is_categorical(col)]
		continuous_columns =  [col for col in columns if not self.is_categorical(col)]

		y = self._obj[label_column].values
		y_p = self._obj[model_prediction_column].values
		x_c = self._obj[continuous_columns].values.astype(float) if len(continuous_columns) > 0 else None
		x_d = self._obj[discrete_columns].values.astype(str) if len(discrete_columns) > 0 else None

		res = regression_model_improvability_analysis(x_c, y_p, y, x_d=x_d, space=space) if problem == 'regression' \
			else classification_model_improvability_analysis(x_c, y_p, y, x_d=x_d, space=space)
		res = res.style.hide_index()

		return res



	def model_explanation_analysis(self, model_prediction_column, input_columns=(), space='dual'):
		"""
		Runs the model explanation analysis on a trained supervised learning model.

		The nature of the supervised learning problem (i.e. regression or classification) is inferred from whether or not :code:`model_prediction_column` is categorical.

		Parameters
		----------
		model_prediction_column : str
			The name of the column containing predicted labels.
		input_columns : set
			List of columns to use as inputs. When an empty set/list is provided, all columns but :code:`model_prediction_column` are used as inputs.
		space : str, 'primal' | 'dual'
			The space in which the maximum entropy problem is solved. 
			When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
			When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.


		Returns
		-------
		res : pandas.Styler
			res.data is a pandas.Dataframe with columns:

				* :code:`'Variable'`: The column name corresponding to the input variable.
				* :code:`'Selection Order'`: The order in which the associated variable was selected, starting at 1 for the most important variable.
				* :code:`'Univariate Explained R^2'`: The :math:`R^2` between predicted labels and this variable.
				* :code:`'Running Explained R^2'`: The :math:`R^2` between predicted labels and all variables selected so far, including this one.
				* :code:`'Marginal Explained R^2'`: The increase in :math:`R^2` between predicted labels and all variables selected so far that is due to adding this variable in the selection scheme.



		.. admonition:: Theoretical Foundation

			Section :ref:`a) Model Explanation`.

		.. seealso::

			* :ref:`kxy.regression.regression_model_explanation_analysis <regression-model-explanation-analysis>`
			* :ref:`kxy.classification.classification_model_explanation_analysis <classification-model-explanation-analysis>`
		"""
		problem = 'classification' if self.is_discrete(model_prediction_column) else 'regression'
		columns = [col for col in self._obj.columns if col != model_prediction_column] if len(input_columns) == 0\
			else input_columns
		discrete_columns = [col for col in columns if self.is_categorical(col)]
		continuous_columns =  [col for col in columns if not self.is_categorical(col)]

		f_x = self._obj[model_prediction_column].values
		x_c = self._obj[continuous_columns].values.astype(float) if len(continuous_columns) > 0 else None
		x_d = self._obj[discrete_columns].values.astype(str) if len(discrete_columns) > 0 else None

		res = regression_model_explanation_analysis(x_c, f_x, x_d=x_d, space=space) if problem == 'regression' \
			else classification_model_explanation_analysis(x_c, f_x, x_d=x_d, space=space)

		variable_columns = continuous_columns + discrete_columns
		res['Variable'] = res['Variable'].map({i: variable_columns[i] for i in range(len(variable_columns))})
		res.set_index(['Variable'], inplace=True)

		cm = sns.light_palette("green", as_cmap=True)
		res = res.style.background_gradient(cmap=cm)

		return res



	def bias(self, bias_source_column, model_prediction_column, linear_scale=True):
		"""
		Quantifies the bias in a supervised learning model as the mutual information between a possible cause and model predictions.

		The nature of the supervised learning problem (i.e. regression or classification) is inferred from whether or not :code:`model_prediction_column` is categorical.


		Parameters
		----------
		bias_source_column : str
			The name of the column containing values of the bias factor (e.g. age, gender, etc.)
		model_prediction_column : str
			The name of the column containing predicted labels associated to the values of :code:`bias_source_column`.
		linear_scale : bool
			Whether the bias should be returned in the linear/correlation scale or in the mutual information scale (in nats).


		Returns
		-------
		 : float
			The mutual information :math:`m` or :math:`1-e^{-2m}` if :code:`linear_scale=True`.


		.. admonition:: Theoretical Foundation

			Section :ref:`b) Quantifying Bias in Models`.

		.. seealso::

			* :ref:`kxy.regression.regression_bias <regression-bias>`
			* :ref:`kxy.classification.classification_bias <classification-bias>`
		"""
		problem = 'classification' if self.is_discrete(model_prediction_column) else 'regression'

		f_x = self._obj[model_prediction_column].values
		z = self._obj[bias_source_column].values

		return regression_bias(f_x, z, linear_scale=linear_scale) if problem == 'regression' \
			else classification_bias(f_x, z, linear_scale=linear_scale)



	def auto_predictability(self, columns=(), space='primal', robust=True, p=None):
		"""
		Estimates the measure of auto-predictability of the time series corresponding to the input columns.

		.. math::

			\\mathbb{PR}\\left(\\{x_t \\} \\right)		:&= h\\left(x_* \\right) - h\\left( \\{ x_t \\} \\right) \\

														 &= h\\left(u_{x_*}\\right) - h\\left( \\{ u_{x_t} \\} \\right).


		Parameters
		----------
		columns : list
			The input columns.
		space : str, 'primal' | 'dual'
			The space in which the maximum entropy problem is solved. 
			When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
			When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.
		p : int, optional
			The number of auto-correlation lags to use as empirical evidence in the maximum-entropy problem. 
			When p is None (the default), it is inferred from the sample.
		p_ic : str
			The criterion used to learn the optimal value of :code:`p` (by fitting a VAR(p) model) when :code:`p=None`. Should be one of 'hqic' (Hannan-Quinn Information Criterion), 'aic' (Akaike Information Criterion), 'bic' (Bayes Information Criterion) and 't-stat' (based on last lag). Same as the 'ic' parameter of :code:`statsmodels.tsa.api.VAR`.


		Returns
		-------
		 : float
			The measure of auto-predictability in nats/period.


		.. note::
			The time series is assumed to have a fixed period, and the index is assumed to be sorted.


		.. seealso::

			* :ref:`kxy.api.core.entropy_rate.auto_predictability <auto-predictability>`
		"""
		if columns == ():
			columns = self._obj.columns
		aps = [(col, auto_predictability(self._obj[col].values, space=space, robust=robust, p=p)) for col in columns]
		aps = sorted(aps, key=lambda x: x[1])
		res = pd.DataFrame({_[0]: [_[1]] for _ in aps})

		return res.T.rename(columns={0: 'PR'})




	def dataset_valuation(self, label_column, existing_input_columns, new_input_columns, space="dual"):
		"""
		Quantifies the additional performance that can be brought about by adding a set of new variables, as the difference between achievable performance with and without the new dataset.


		Parameters
		----------
		label_column : str
			The name of the column containing the true labels of the supervised learning problem.
		existing_input_columns : set
			List of columns corresponding to existing inputs/variables.
		new_input_columns : set
			List of columns corresponding to the new dataset.
		space : str, 'primal' | 'dual'
			The space in which the maximum entropy problem is solved. 
			When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
			When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.


		Returns
		-------
		res : pandas.Styler
			res.data is a Dataframe with columns (where applicable):

			* :code:`'Increased Achievable R^2'`: The highest :math:`R^2` increase that can result from adding the new dataset.
			* :code:`'Increased Achievable Log-Likelihood Per Sample'`: The highest increase in the true log-likelihood per sample that can result from adding the new dataset.
			* :code:`'Increased Achievable Accuracy'`: The highest increase in the classification accuracy that can result from adding the new dataset.


		.. admonition:: Theoretical Foundation

			Section :ref:`1 - Achievable Performance`.

		.. seealso::

			* :ref:`kxy.regression.regression_achievable_performance_analysis <regression-achievable-performance-analysis>`
			* :ref:`kxy.classification.classification_achievable_performance_analysis <classification-achievable-performance-analysis>`
		"""
		if existing_input_columns is None or len(existing_input_columns) == 0:
			return self.achievable_performance_analysis(label_column, input_columns=new_input_columns)

		if new_input_columns is None or len(new_input_columns) == 0:
			problem = 'classification' if self.is_discrete(label_column) else 'regression'
			if problem == 'classification':
				probas = [1.*(self._obj[self._obj[label_column]==cat].shape[0])/self._obj.shape[0] for cat in list(set(list(self._obj[label_column].values)))]
				return pd.DataFrame({\
					'Achievable R^2': [0.0], \
					'Achievable Log-Likelihood Per Sample': [0.0], \
					'Achievable Accuracy': [np.max(probas)]})

			else:
				return pd.DataFrame({\
					'Achievable R^2': [0.0], \
					'Achievable Log-Likelihood Per Sample': [0.0]})	

		all_inputs = set(list(existing_input_columns)+list(new_input_columns))
		new_perf = self.achievable_performance_analysis(label_column, input_columns=all_inputs, space=space)
		old_perf = self.achievable_performance_analysis(label_column, input_columns=existing_input_columns, space=space)
		imp_perf = new_perf-old_perf
		imp_perf.rename(columns={col: col.replace('Achievable', 'Increased Achievable') for col in imp_perf.columns}, inplace=True)

		cm = sns.light_palette("green", as_cmap=True)
		imp_perf = imp_perf.style.background_gradient(cmap=cm)

		return imp_perf



	def __hash__(self):
		return hash(self._obj.to_string())






#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging
import requests
from time import sleep, time

import numpy as np

from .client import APIClient



def solve_copula_async(corr):
	"""
	.. _solve-copula-async:
	Solve the maximum-entropy copula problem under Spearman rank correlation matrix constraints asynchronously.

	.. note:: 

		The solution to the optimization problem is not returned. This function should be used to pre-compute the solution to the optimization problem for later use.


	.. seealso::

		:ref:`kxy.api.optimizers.solve_sync <solve-copula-sync>`.


	Parameters
	----------
	corrr : np.array
		The Spearman correlation matrix.


	Returns
	-------
	success : bool
		Whether the request went through successfully.
	"""
	try:
		c = json.dumps([['%.3f' % corr[i, j]  for j in range(corr.shape[1])] for i in range(corr.shape[0])])
		logging.debug('Launching a max-ent solver in the background with corr=%s' % c)
		api_response = APIClient.route(path='/core/dependence/copula/maximum-entropy/entropy/rv/pre-compute', \
			method='POST', corr=c)

		if api_response.status_code == requests.codes.ok:
			return True

		logging.info(api_response.json())
		return False

	except:
		return False





def solve_copula_sync(corr, mode=None, output_index=None, solve_async=True, space='dual'):
	"""
	.. _solve-copula-sync:
	Solve the maximum-entropy copula problem under Spearman rank correlation matrix constraints synchronously.

	.. note:: 

		This function blocks until the optimization problem is solved, and returrns the requested quantity.

	.. seealso::

		:ref:`kxy.api.optimizers.solve_sync <solve-copula-async>`.


	Parameters
	----------
	corrr : np.array
		The Spearman correlation matrix.

	mode : str
		One of :code:`'copula_entropy'`, and :code:`'mutual_information_v_output'`.

		When mode is :code:`'copula_entropy'` the function returns the entropy of the copula of the random
		vector whose Spearman correlation matrix is the input :code:`corr`.

		When mode is :code:`'mutual_information_v_output'` the function returns the mutual information between 
		a continuous output variable :math:`y` (specificied by :code:`output_index`) and continuous input 
		variables :math:`x` (specificied by the other variables): :math:`I(x, y)`. We recall that the mutual information 
		between two continuous variables is the mutual information of their copula-uniform dual representations. The copula 
		of :math:`(x, y)` is learned as the maximum-entropy copula under the constraint that the Spearman correlation matrix 
		is the input :code:`corr`.

	output_index : int
		The index of the column that should be used as output variable when mode is :code:`'mutual_information_v_output'`.



	Returns
	-------
	r : float
		The requested result, namely the copula entropy, the mutual information or the conditional
		mutual information, depending on the value of the mode.
	"""
	assert mode in ('copula_entropy', 'mutual_information_v_output')

	if mode in ['mutual_information_v_output', 'conditional_mutual_information']:
		assert output_index is not None, 'The output index should be provided'

	if solve_async:
		solve_copula_async(corr)

	if mode == 'mutual_information_v_output':
		res = mutual_information_analysis(corr, output_index, space=space)
		if res is None:
			return None

		return res['mutual_information']

	if mode == 'copula_entropy':
		res = copula_entropy_analysis(corr, space=space)
		if res is None:
			return None
		return res['copula_entropy']


	return None



def mutual_information_analysis(corr, output_index, space='dual', greedy=True):
	'''
	Analyzes the dependency between :math:`d`-dimensional continuous random vector :math:`x=\\left(x_1, \\dots, x_d \\right)` and
	a continuous random scalar :math:`y` whose joint correlation matrix is :code:`corr`, the column :code:`output_index` of which 
	represents the variable :math:`y` and the others the variable :math:`x`.


	Recall that, for any permutation :math:`\\pi_1, \\dots, \\pi_d` of :math:`1, \\dots, d`, by the tower law,

	.. math::

		I\\left(y; x_1, \\dots, x_d\\right) = I\\left(y; x_{\\pi_1}\\right) + \\sum_{i=2}^d I\\left(y; x_{\\pi_i} \\vert x_{\\pi_{i-1}}, \\dots, x_{\\pi_1} \\right).


	This function estimates the mutual information :math:`I(y; x)` by learning the following permutation. 

	When greedy is True:

	* :math:`x_{\\pi_1}` is the input with the largest maximum entropy mutual information with :math:`y` under Spearman rank correlation constraints.

	* :math:`x_{\\pi_i}` for :math:`i>1` is the input with the largest maximum entropy conditional mutual information :math:`I\\left(y; * \\vert x_{\\pi_{i-1}}, \\dots, x_{\\pi_1}\\right)`. Note that by the time :math:`\\pi_i` is selected, :math:`I\\left(y; x_{\\pi_{i-1}}, \\dots, x_{\\pi_1}\\right)` is already known, so that the maximum entropy conditional mutual information is simply derived from the maximum entropy copula distribution of :math:`I\\left(y; x_{\\pi_i}, \\dots, x_{\\pi_1}\\right)`.

	This function returns the learned permutation of inputs, the associated conditional mutual informations (a.k.a, the incremental input importance scores), as well as the mutual information :math:`I\\left(y; x_1, \\dots, x_d\\right)`.

	When greedy is False, :math:`\\pi_i` is the input with the i-th largest mutual information with the output.


	Parameters
	----------
	corrr : np.array
		The Spearman correlation matrix.

	output_index: int
		The index of the column to use as output.


	Returns
	-------
	res : dict
		Dictionary with keys :code:`mutual_information`, :code:`selection_order`, and :code:`conditional_mutual_informations`.
	'''
	c = json.dumps([['%.3f' % corr[i, j]  for j in range(corr.shape[1])] for i in range(corr.shape[0])])
	opt_launched = False
	max_retry = 60
	first_try = True
	retry_count = 0
	request_id = ''
	while (first_try or api_response.status_code == requests.codes.retry) and retry_count < max_retry:
		first_try = False

		if request_id == '':
			query_start_time = time()
			# First attempt
			logging.debug('Querying mutual information analysis with corr=%s and output_index=%d' % (c, output_index))
			api_response = APIClient.route(path='/rv/mutual-information-analysis', method='POST',\
				corr=c, output_index=output_index, request_id=request_id, timestamp=int(time()), \
				space=space, greedy=int(greedy))
			query_duration = time()-query_start_time

		else:
			query_start_time = time()
			# Subsequent attempt: refer to the initial request
			logging.debug('Querying mutual information analysis for request_id=%s' % request_id)
			api_response = APIClient.route(path='/rv/mutual-information-analysis', method='POST',\
				request_id=request_id, timestamp=int(time()), space=space)
			query_duration = time()-query_start_time

		retry_count += 1
		if api_response.status_code == requests.codes.retry:
			request_id = api_response.json()['request_id']
			sleep(.1 if query_duration > 10. else 10.)


	if api_response.status_code == requests.codes.ok:
		return api_response.json()

	else:
		logging.warning(api_response.json())

	return None



def copula_entropy_analysis(corr, space='dual'):
	'''
	Analyzes the entropy of the copula of a :math:`d`-dimensional continuous random vector :math:`x=\\left(x_1, \\dots, x_d \\right)`, with copula-uniform representation :math:`u=\\left(u_1, \\dots, u_d \\right)`.

	Recall that, for any permutation :math:`(1), \\dots, (d)` of :math:`1, \\dots, d`, by the tower law,

	.. math::

		h\\left(u_1, \\dots, u_d\\right) = \\sum_{i=2}^d h\\left( u_{(i)} \\vert u_{(i-1)}, \\dots, u_{(1)} \\right).


	This function estimates the copula entropy  :math:`h(u)` by learning the following permutation:

	* :math:`x_{(1)}` and :math:`x_{(2)}` are chosen to be the two random variables with smallest copula entropy (or equivalently, the highest mutual information).

	* :math:`x_{(i)}` for :math:`i>1` is the input with the smallest conditional copula entropy :math:`h\\left(* \\vert u_{(i-1)}, \\dots, u_{(1)} \\right)` (or equivalently, the highest mutual information :math:`I\\left(*; x_{(1)}, \\dots, x_{(i-1)}\\right)`). Note that by the time :math:`(i)` is selected, :math:`h\\left(u_{(i-1)}, \\dots, u_{(1)}\\right)` is already known, so that the maximum entropy conditional entropy is simply derived from the maximum entropy copula distribution of :math:`\\left(x_{(i)}, \\dots, x_{(1)}\\right)`.

	This function returns the learned permutation of inputs, the association conditional entropies, as well as the copula entropy :math:`h(u)`.


	Parameters
	----------
	corrr : np.array
		The Spearman correlation matrix.

	output_index: int
		The index of the column to use as output.


	Returns
	-------
	res : dict
		Dictionary with keys :code:`copula_entropy`, :code:`selection_order`, and :code:`conditional_copula_entropies`.
	'''
	c = json.dumps([['%.3f' % corr[i, j]  for j in range(corr.shape[1])] for i in range(corr.shape[0])])
	opt_launched = False
	max_retry = 60
	first_try = True
	retry_count = 0
	request_id = ''
	while (first_try or api_response.status_code == requests.codes.retry) and retry_count < max_retry:
		first_try = False

		if request_id == '':
			query_start_time = time()
			# First attempt
			logging.debug('Querying copula entropy analysis with corr=%s' % (c))
			api_response = APIClient.route(path='/rv/copula-entropy-analysis', method='POST', corr=c, \
				request_id=request_id, timestamp=int(time()), space=space)
			query_duration = time()-query_start_time

		else:
			query_start_time = time()
			# Subsequent attempt: refer to the initial request
			logging.debug('Querying copula entropy analysis for request_id=%s' % request_id)
			api_response = APIClient.route(path='/rv/copula-entropy-analysis', method='POST', \
				request_id=request_id, timestamp=int(time()), space=space)
			query_duration = time()-query_start_time

		retry_count += 1
		if api_response.status_code == requests.codes.retry:
			request_id = api_response.json()['request_id']
			sleep(.1 if query_duration > 10. else 10.)


	if api_response.status_code == requests.codes.ok:
		return api_response.json()

	else:
		logging.warning(api_response.json())

	return None




def information_adjusted_correlation_from_spearman(corr, space='dual'):
	'''
	Determine the Pearson correlation matrix that should be used as plug-in replacement
	for the standard Pearson correlation estimator, in situations where Gaussianity has
	to be assumed for simplicity, but the data-generating distribution exhibits non-Gaussian
	traits (e.g. heavier or lighter tails).

	We proceed as follows, for every pairwise Spearman rank correlation, we estmate the smallest
	mutual information that is consistent with the pairwise Spearman rank correlation. This is also 
	the mutual information of the model which, among all models that have the same Spearman rank 
	correlation, is the most uncertain about everything else. 

	This mutual information captures the essence of the dependence between the two variables without
	assuming Gaussianity. To go back to the Gaussian scale, we determine the Pearson correlation which, 
	under the Gaussian assumption would yield the same mutual information as estimated.


	Parameters
	----------
	corr : np.array
		The Spearman correlation matrix.


	Returns
	-------
	res : np.array
		The array of equivalent Pearson correlation coefficients.
	'''
	c = json.dumps([['%.3f' % corr[i, j]  for j in range(corr.shape[1])] for i in range(corr.shape[0])])
	opt_launched = False
	max_retry = 60
	first_try = True
	retry_count = 0
	request_id = ''
	while (first_try or api_response.status_code == requests.codes.retry) and retry_count < max_retry:
		first_try = False

		if request_id == '':
			query_start_time = time()
			# First attempt
			logging.debug('Querying information adjusted correlation with corr=%s' % (c))
			api_response = APIClient.route(path='/rv/information-adjusted-correlation', method='POST', corr=c, \
				request_id=request_id, timestamp=int(time()), space=space)
			query_duration = time()-query_start_time

		else:
			query_start_time = time()
			# Subsequent attempt: refer to the initial request
			logging.debug('Querying information adjusted correlation for request_id=%s' % request_id)
			api_response = APIClient.route(path='/rv/information-adjusted-correlation', method='POST', \
				request_id=request_id, timestamp=int(time()), space=space)
			query_duration = time()-query_start_time

		retry_count += 1
		if api_response.status_code == requests.codes.retry:
			request_id = api_response.json()['request_id']
			sleep(.1 if query_duration > 10. else 10.)


	if api_response.status_code == requests.codes.ok:
		return np.array(api_response.json()['ia_corr']).astype(float)

	else:
		logging.warning(api_response.json())

	return None



def robust_pearson_corr_from_spearman(corr):
	'''
	Return the Pearson correlation matrix that is equivalent to the input 
	Spearman rank correlation matrix, assuming inputs are jointly Gaussian.


	Parameters
	----------
	corr : np.array
		The Spearman correlation matrix.


	Returns
	-------
	res : np.array
		The array of equivalent Pearson correlation coefficients.
	'''
	return information_adjusted_correlation_from_spearman(corr, space='primal')






#!/usr/bin/env python
# -*- coding: utf-8 -*-
from concurrent.futures import ThreadPoolExecutor

import numpy as np
import pandas as pd

from kxy.api.core import least_mixed_mutual_information, discrete_mutual_information, \
	discrete_entropy, hqi, least_mixed_conditional_mutual_information


def classification_achievable_performance_analysis(x_c, y, x_d=None, space='dual'):
	"""
	.. _classification-achievable-performance-analysis:
	Quantifies the performance that can be achieved when trying to predict :math:`y` with :math:`x`.

	Parameters
	----------
	x_c : (n,d) np.array
		Continuous inputs.
	y : (n,) np.array
		Labels.
	x_d : (n, d) np.array or None (default), optional
		Discrete inputs.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.


	Returns
	-------
	a : pandas.DataFrame

		Dataframe with columns:

		* :code:`'Achievable R^2'`: The highest :math:`R^2` that can be achieved by a classification model using provided inputs to predict the label.
		* :code:`'Achievable Log-Likelihood Per Sample'`: The highest true log-likelihood per sample that can be achieved by a classification model using provided inputs to predict the label.
		* :code:`'Achievable Accuracy'`: The highest classification accuracy that can be achieved by a classification model using provided inputs to predict the label.


	.. admonition:: Theoretical Foundation

		Section :ref:`1 - Achievable Performance`.
	"""
	assert x_d is not None or x_c is not None, "x_c and x_d cannot be both None."

	if x_c is None:
		mi = discrete_mutual_information(x_d, y)

	else:
		mi = least_mixed_mutual_information(x_c, y, x_d=x_d, space=space, non_monotonic_extension=True)

	hy = discrete_entropy(y)
	q = len(list(set(y)))
	n = y.shape[0]

	return pd.DataFrame({\
		'Achievable R^2': [1.-np.exp(-2.*mi)], \
		'Achievable Log-Likelihood Per Sample': [-max(hy-mi, 0.0)], \
		'Achievable Accuracy': [hqi(max(hy-mi, 0.0), q)]})



def classification_variable_selection_analysis(x_c, y, x_d=None, space='dual'):
	"""
	.. _classification-variable-selection-analysis:
	Runs the variable selection analysis for a classification problem.

	Parameters
	----------
	x_c : (n, d) np.array
		Continuous inputs.
	y : (n,) np.array
		Labels.
	x_d : (n, q) np.array or None (default), optional
		Discrete inputs.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.


	Returns
	-------
	a : pandas.DataFrame

		Dataframe with columns:

		* :code:`'Variable'`: The variable index starting from 0 at the leftmost column of :code:`x_c` and ending at the rightmost column of :code:`x_d`.
		* :code:`'Selection Order'`: The order in which the associated variable was selected, starting at 1 for the most important variable.
		* :code:`'Univariate Achievable R^2'`: The highest :math:`R^2` that can be achieved by a classification model solely using this variable.
		* :code:`'Maximum Marginal R^2 Increase'`: The highest amount by which the :math:`R^2` can be increased as a result of adding this variable in the variable selection scheme.
		* :code:`'Running Achievable R^2'`: The highest :math:`R^2` that can be achieved by a classification model using all variables selected so far, including this one.
		* :code:`'Univariate Achievable Accuracy'`: The highest classification accuracy that can be achieved by a classification model solely using this variable.
		* :code:`'Maximum Marginal Accuracy Increase'`: The highest amount by which the classification accuracy can be increased as a result of adding this variable in the variable selection scheme.
		* :code:`'Running Achievable Accuracy'`: The highest classification accuracy that can be achieved by a classification model using all variables selected so far, including this one.
		* :code:`'Conditional Mutual Information (nats)'`: The mutual information between this variable and the label, conditional on all variables previously selected.
		* :code:`'Running Mutual Information (nats)'`: The mutual information between all variables selected so far, including this one, and the label.
		* :code:`'Univariate Achievable True Log-Likelihood Per Sample'`: The highest true log-likelihood per sample that can be achieved by a classification model solely using this variable.
		* :code:`'Maximum Marginal True Log-Likelihood Per Sample Increase'`: The highest amount by which the true log-likelihood per sample can be increased as a result of adding this variable in the variable selection scheme.
		* :code:`'Running Achievable True Log-Likelihood Per Sample'`: The highest true log-likelihood per sample that can be achieved by a classification model using all variables selected so far, including this one.
		* :code:`'Maximum Marginal True Log-Likelihood Increase Per Sample'`: The highest amount by which the true log-likelihood per sample can increase as a result of adding this variable.
		* :code:`'Running Maximum Log-Likelihood Increase Per Sample'`: The highest amount by which the true log-likelihood per sample can increase (over the log-likelihood of the naive strategy consisting of predicting the mode of :math:`y`) as a result of using all variables selected so far, including this one.


	.. admonition:: Theoretical Foundation

		Section :ref:`2 - Variable Selection Analysis`.
	"""
	x_c_ = x_c[:, None] if len(x_c.shape) == 1 else x_c
	x_d_ = None if x_d is None else x_d[:, None] if len(x_d.shape) == 1 else x_d
	n_inputs = x_c_.shape[1] if x_d is None else x_c_.shape[1] + x_d_.shape[1]
	inputs = [_ for _ in range(n_inputs)]
	categorical_inputs = [_ for _ in range(x_c_.shape[1], n_inputs)]

	final_cmis = {}
	final_mis = {}
	final_univariate_rsq = {}
	final_univariate_acc = {}
	final_univariate_llik = {}
	order = {}
	categorical_conditions = []
	continuous_conditions = []
	hy = discrete_entropy(y)
	q = len(list(set(y)))
	n = y.shape[0]

	def conditional_mutual_information(args):
		# Calculate the conditional mutual information
		i_, cont_cs, cat_cs = args
		_cat_cs = [_-x_c_.shape[1] for _ in cat_cs]
		z_c = None if len(cont_cs) == 0 else x_c_[:, list(cont_cs)]
		z_d = None if len(cat_cs) == 0 else x_d_[:, list(_cat_cs)]
		_x_c = None if i_ in categorical_inputs else x_c_[:, [i_]]
		_x_d = None if x_d_ is None else x_d_[:, i_-x_c_.shape[1]] if i_ in categorical_inputs else None
		is_cat = i_ in categorical_inputs

		mi_ = discrete_mutual_information(_x_d, y) if _x_c is None else \
			least_mixed_mutual_information(_x_c, y, space=space, non_monotonic_extension=True)

		mi_ = min(mi_, hy) # I(y, x) = h(y)-h(y|x) <= h(y) when y is discrete.

		if len(cont_cs) == 0 and len(cat_cs) == 0:
			# (discrete, continuous) | None, or (discrete, discrete) | None
			cmi_ = mi_

		else:
			# (discrete, continuous | discrete & continuous)
			cmi_ = least_mixed_conditional_mutual_information(_x_c, y, z_c, x_d=_x_d, z_d=z_d, \
				space=space, non_monotonic_extension=True)
			cmi_ = min(cmi_, hy)
			
		return {i_: [cmi_, mi_]}


	while len(inputs) > 0:
		cmis = {}
		with ThreadPoolExecutor(max_workers=10) as p:
			args = [(i, continuous_conditions, categorical_conditions) for i in inputs]
			for cmi in p.map(conditional_mutual_information, args):
				cmis.update(cmi)

		for key, value in sorted(cmis.items(), key=lambda x: -x[1][0]):
			final_cmis[key] = value[0]
			final_mis[key] = value[1]
			final_univariate_rsq[key] = 1.-np.exp(-2.*final_mis[key])
			final_univariate_acc[key] = hqi(max(hy-final_mis[key], 0.0), q)
			final_univariate_llik[key] = min(final_mis[key]-hy, 0.0)

			inputs.remove(key)
			if key in categorical_inputs:
				categorical_conditions += [key]
			else:
				continuous_conditions += [key]
			order[key] = len(categorical_conditions) + len(continuous_conditions)
			break


	final_rsq_inc = {}
	final_run_rsqs = {}

	final_acc_inc = {}
	final_run_acc = {}

	final_llik_inc = {}
	final_run_llik = {}

	run_mi = 0
	for v, o in sorted(order.items(), key=lambda item: item[1]):
		run_mi += final_cmis[v]
		final_run_rsqs[o] = 1.-np.exp(-2.*run_mi)
		final_run_acc[o] = hqi(max(hy-run_mi, 0.0), q)
		final_run_llik[o] = -max(hy-run_mi, 0.0)

		if final_rsq_inc == {}:
			final_rsq_inc[o] = final_run_rsqs[o]
			final_acc_inc[o] = final_run_acc[o]
			final_llik_inc[o] = final_run_llik[o]
			
		else:
			final_rsq_inc[o] = final_run_rsqs[o]-final_run_rsqs[o-1]
			final_acc_inc[o] = final_run_acc[o]-final_run_acc[o-1]
			final_llik_inc[o] = final_run_llik[o]-final_run_llik[o-1]


	sorted_order = sorted(order.items(), key=lambda item: item[1])
	max_r_2 = np.max([_ for _ in final_run_rsqs.values()])
	importance_df = pd.DataFrame({
		'Variable': [v for v, o in sorted_order], \
		'Selection Order': [o for v, o in sorted_order], \

		# Achievable R^2 analysis
		'Univariate Achievable R^2': [final_univariate_rsq[v] for v, o in sorted_order], \
		'Maximum Marginal R^2 Increase': [final_rsq_inc[o] for v, o in sorted_order], \
		'Running Achievable R^2': [final_run_rsqs[o] for v, o in sorted_order], \
		'Running Achievable R^2 (%)': [100.*final_run_rsqs[o]/max_r_2 for v, o in sorted_order], \

		# Achievable accuracy analysis
		'Univariate Achievable Accuracy': [final_univariate_acc[v] for v, o in sorted_order], \
		'Maximum Marginal Accuracy Increase': [final_acc_inc[o] for v, o in sorted_order], \
		'Running Achievable Accuracy': [final_run_acc[o] for v, o in sorted_order], \

		# Achievable log-likelihood analysis
		'Univariate Achievable True Log-Likelihood Per Sample': [final_univariate_llik[v] for v, o in sorted_order], \
		'Maximum Marginal True Log-Likelihood Per Sample Increase': [final_llik_inc[o] for v, o in sorted_order], \
		'Running Achievable True Log-Likelihood Per Sample': [final_run_llik[o] for v, o in sorted_order], \

		# Conditional mutual information analysis
		'Univariate Mutual Information (nats)': [final_mis[v] for v, o in sorted_order], \
		'Conditional Mutual Information (nats)': [final_cmis[v] for v, o in sorted_order]})
	importance_df['Running Mutual Information (nats)'] = importance_df['Conditional Mutual Information (nats)'].cumsum()
	max_mi = importance_df['Running Mutual Information (nats)'].max()
	importance_df['Running Mutual Information (%)'] = 100.*importance_df['Running Mutual Information (nats)']/max_mi

	return importance_df





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np 
from numpy.dual import svd

def spearman_corr(x):
	"""
	Calculate the Spearman rank correlation matrix, ignoring nans.

	Parameters
	----------
	x : (n, d) np.array
		Input data representing n i.i.d. draws from the d-dimensional 
		random variable, whose Spearman rank correlation matrix this 
		function calculates.


	Returns
	-------
	corr : np.array
		The Spearman rank correlation matrix.
	"""
	n, m = x.shape
	mask = np.isnan(x).copy()
	valid_mask = np.logical_not(mask).astype(int)

	# R[i,j] is the rank of x[i, j] among x[1, j] ... x[n, j]
	R = (1+x.argsort(axis=0).argsort(axis=0)).astype(float)
	np.copyto(R, np.nan, where=mask)
	o = np.ones((m, m))
	non_nan_ns = np.einsum('ij, ik->jk', valid_mask, valid_mask) # Number of non-nan pairs
	mean_R = np.nanmean(R, axis=0)
	demeaned_R = R - mean_R
	standard_R = demeaned_R/np.nanstd(demeaned_R, axis=0)
	np.copyto(standard_R, 0.0, where=mask)
	corr = np.divide(np.einsum('ij, ik->jk', standard_R, standard_R), non_nan_ns)
	corr = np.round(corr, 3)

	return corr



def pearson_corr(x):
	"""
	Calculate the Pearson correlation matrix, ignoring nans.

	Parameters
	----------
	x : (n, d) np.array
		Input data representing n i.i.d. draws from the d-dimensional 
		random variable, whose Pearson correlation matrix this 
		function calculates.


	Returns
	-------
	corr : np.array
		The Pearson correlation matrix.
	"""
	n, m = x.shape
	mask = np.isnan(x).copy()
	valid_mask = np.logical_not(mask).astype(int)

	np.copyto(x, np.nan, where=mask)
	o = np.ones((m, m))
	non_nan_ns = np.einsum('ij, ik->jk', valid_mask, valid_mask) # Number of non-nan pairs
	mean_x = np.nanmean(x, axis=0)
	demeaned_x = x - mean_x
	standard_x = demeaned_x/np.nanstd(demeaned_x, axis=0)
	np.copyto(standard_x, 0.0, where=mask)
	corr = np.divide(np.einsum('ij, ik->jk', standard_x, standard_x), non_nan_ns)
	corr = np.round(corr, 3)

	return corr




def robust_log_det(c):
	"""
	Computes the logarithm of the determinant of a positive definite matrix in a fashion that is more robust to ill-conditioning than taking the logarithm of np.linalg.det.

	.. note::
		Specifically, we compute the SVD of c, and return the sum of the log of eigenvalues. np.linalg.det on the other hand computes the Cholesky decomposition of c, which is more likely to fail than its SVD, and takes the product of its diagonal elements, which could be subject to underflow error when diagonal elements are small.

	Parameters
	----------
	c: (d, d) np.array 
		Square input matrix for computing log-determinant.

	Returns
	-------
	d : float
		Log-determinant of the input matrix.
	"""
	u, s, v = svd(c)
	
	return np.sum(np.log(s))


def hqi(h, q):
	"""
	Computes :math:`\\bar{h}_q^{-1}(x)` where

	.. math::

		\\bar{h}_q(a) = -a \\log a -(1-a) \\log \\left(\\frac{1-a}{q-1}\\right), ~~~~ a \\geq \\frac{1}{q}.
	"""
	if h==0:
		return 1.0

	a = np.round(np.arange(1./q, 1.0, 0.001), 3)
	h_ = -a*np.log(a)-(1.-a)*np.log((1.-a)/(q-1.))
	idx = np.abs(h_-h).argmin()

	return a[idx]





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import numpy as np
import requests

from kxy.api import APIClient, solve_copula_sync

from .utils import spearman_corr, pearson_corr
from .entropy import least_structured_copula_entropy, least_structured_continuous_entropy, \
	least_structured_mixed_entropy, discrete_entropy


def least_total_correlation(x, space='dual'):
	""" 
	.. _least-total-correlation:
	Estimates the smallest total correlation of a continuous d-dimensional random vector,
	that is consistent with the average pairwise Spearman rank correlation estimated 
	from the input sample.

	.. math::
		TC(x) := \sum_{i=1}^d h(x_i)-h(x) = -h(u).

	This is the negative of :ref:`kxy.api.core.entropy.least_structured_copula_entropy <least-structured-copula-entropy>`.

	Parameters
	----------
	x : (n, d) np.array
		n i.i.d. draws from the data generating distribution.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.

	Returns
	-------
	tc : float
		The smallest total correlation consistent with observed empirical evidence, in nats.

	Raises
	------
	AssertionError
		If x is not a two dimensional array.
	"""
	assert len(x.shape) == 2
	return -least_structured_copula_entropy(x, space=space)



def least_continuous_mutual_information(x, y, space='dual', non_monotonic_extension=True):
	"""
	.. _least-continuous-mutual-information:
	Estimates the mutual information between a :math:`d`-dimensional random vector :math:`x` of inputs
	and a continuous scalar random variable :math:`y` (or label), assuming the least amount of 
	structure in :math:`x` and :math:`(x, y)`, other than what is necessary to be consistent with some 
	observed properties.

	.. note::

		.. math::
			I(x, y) &= h(x) + h(y) - h(x, y) \\

			        &= h\\left(u_x\\right) - h\\left(u_{x,y}\\right)
		
		where :math:`u_x` (resp. :math:`u_{x,y}`) is the copula-uniform representation of :math:`x`
		(resp. :math:`(x, y)`).

		We use as model for :math:`u_{x, y}` the least structured copula that is consistent with 
		the Spearman rank correlation matrix of the joint vector :math:`(x, y)`. 

	Parameters
	----------
	x : (n, d) np.array
		n i.i.d. draws from the features generating distribution.
	y : (n,) np.array
		n i.i.d. draws from the (continuous) labels generating distribution, sampled
		jointly with x.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.

	Returns
	-------
	i : float
		The mutual information between x and y, **in nats**.

	Raises
	------
	AssertionError
		If y is not a one dimensional array or x or y are not numeric.
	"""
	assert np.can_cast(x, float) and np.can_cast(y, float), 'x and y should represent draws from continuous random variables.'
	assert len(y.shape) == 1 or y.shape[1] == 1, 'Only one-dimensional outputs are supported for now.'

	x_ = np.reshape(x, (len(x), 1)) if len(x.shape) == 1 else x.copy()
	y_ = np.reshape(y, (len(y), 1)) if len(y.shape) == 1 else y.copy()

	corr = pearson_corr(np.hstack([y_, x_])) if space == 'primal' or not non_monotonic_extension else \
		spearman_corr(np.hstack([y_, x_, np.abs(x_-x_.mean(axis=0))]))
	mi = solve_copula_sync(corr, mode='mutual_information_v_output', output_index=0, solve_async=False, \
		space=space)

	return mi



def least_continuous_conditional_mutual_information(x, y, z, space='dual', non_monotonic_extension=True):
	"""
	.. _least-continuous-conditional-mutual-information:
	Estimates the conditional mutual information between a :math:`d`-dimensional random vector :math:`x` of inputs
	and a continuous scalar random variable :math:`y` (or label), conditional on a third continuous random 
	variable :math:`z`, assuming the least amount of structure in :math:`(x, y, z)`, other than what is 
	necessary to be consistent with some observed properties.

	.. note::

		.. math::
			I(x; y|z) &= h(x|z) + h(y|z) - h(x, y|z) \\

			        &= h\\left(u_x | u_z \\right) - h\\left(u_{x,y} | u_z \\right)

			        &= I(y; x, z) - I(y; z)

		
		where :math:`u_x` (resp. :math:`u_{x,y}`, :math:`u_z`) is the copula-uniform representation of :math:`x`
		(resp. :math:`(x, y)`, :math:`z`).

		We use as model for :math:`u_{x, y, z}` the least structured copula that is consistent with 
		maximum entropy constraints in the chosen space. 

	Parameters
	----------
	x : (n, d) np.array
		n i.i.d. draws from the inputs generating distribution.
	y : (n,) np.array
		n i.i.d. draws from the (continuous) labels generating distribution, sampled
		jointly with x.
	z : (n,) np.array
		n i.i.d. draws from the (continuous) labels generating distribution, sampled
		jointly with z.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.

	Returns
	-------
	i : float
		The mutual information between x and y, **in nats**.

	Raises
	------
	AssertionError
		If y is not a one dimensional array or x, y and z are not all numeric.
	"""
	assert np.can_cast(x, float) and np.can_cast(y,  float) and np.can_cast(z, float), \
		'x, y and z should represent draws from continuous random variables.'
	assert len(y.shape) == 1 or y.shape[1] == 1, 'Only one-dimensional outputs are supported for now.'

	x_ = np.reshape(x, (len(x), 1)) if len(x.shape) == 1 else x.copy()
	y_ = np.reshape(y, (len(y), 1)) if len(y.shape) == 1 else y.copy()
	z_ = np.reshape(z, (len(z), 1)) if len(z.shape) == 1 else z.copy()

	mi_y_xz = least_continuous_mutual_information(np.hstack([x_, z_]), y_, space=space, \
		non_monotonic_extension=non_monotonic_extension)
	mi_y_z = least_continuous_mutual_information(z_, y_, space=space, \
		non_monotonic_extension=non_monotonic_extension)

	return max(mi_y_xz-mi_y_z, 0.0)




def least_mixed_mutual_information(x_c, y, x_d=None, space='dual', non_monotonic_extension=True):
	"""
	.. _least-mixed-mutual-information:
	Estimates the mutual inforrmation between some features (a d-dimensional random vector
	:math:`x_c` and possibly a discrete feature variable :math:`x_d` and a discrete output/label 
	random variable.	

	.. note::

		.. math::
			I({x_c, x_d}; y) &= h(y) + h(x_c, x_d) - h(x_c, x_d, y) \\

							 &= h(x_c, x_d) - h\\left(x_c, x_d \\vert y\\right) \\

							 &= h(x_c, x_d) - E\\left[h\\left(x_c, x_d \\vert y=.\\right)\\right] \\

							:&= h(x_c, x_d) - \sum_{j=1}^q \mathbb{P}(y=j)  h\\left(x_c, x_d \\vert y=j \\right)

		
		When there are discrete features:

		* :math:`h(x_c, x_d)` is estimated using :ref:`kxy.api.core.entropy.least_structured_mixed_entropy <least-structured-mixed-entropy>`,
		* :math:`\mathbb{P}(y=j)` is estimated using frequencies, 
		* :math:`h\\left(x_c, x_d | y=j\\right)` is estimated using :ref:`kxy.api.core.entropy.least_structured_mixed_entropy <least-structured-mixed-entropy>`.

		When there are no discrete features, :math:`x_d` can simply be removed from the equations above, and :ref:`kxy.api.core.entropy.least_structured_continuous_entropy <least-structured-continuous-entropy>` is used for estimation instead of :ref:`kxy.api.core.entropy.least_structured_mixed_entropy <least-structured-mixed-entropy>`.

	Parameters
	----------
	x_c : (n, d) np.array
		n i.i.d. draws from the continuous data generating distribution.
	x_d : (n,) np.array or None (default)
		n i.i.d. draws from the discrete data generating distribution, jointly sampled with x_c, or None
		if there are no discrete features.
	y : (n,) np.array
		n i.i.d. draws from the (discrete) labels generating distribution, sampled jointly with x.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.

	Returns
	-------
	i : float
		The mutual information between features and discrete labels, **in nats**.

	Raises
	------
	AssertionError
		If y is not a one-dimensional array, or x_c is not an array of numbers.
	"""
	assert np.can_cast(x_c, float), 'x_c should represent draws from a continuous random variable.'
	assert len(y.shape) == 1 or y.shape[1] == 1, 'Only one-dimensional outputs are supported for now.'

	x_c_ = np.reshape(x_c, (len(x_c), 1)) if len(x_c.shape) == 1 else x_c.copy()
	x_c_r = np.hstack((x_c_, np.abs(x_c_-x_c_.mean(axis=0)))) if non_monotonic_extension else x_c_

	categories = list(set(list(y)))
	n = y.shape[0]
	probas = np.array([1.*len(y[y==cat])/n for cat in categories])

	h = least_structured_continuous_entropy(x_c_r, space=space) if x_d is None else least_structured_mixed_entropy(x_c_r, x_d, space=space)
	wh = np.sum([probas[i] * least_structured_continuous_entropy(x_c_r[y==categories[i]], space=space) for i in range(len(categories))]) if x_d is None \
		else np.sum([probas[i] * least_structured_mixed_entropy(x_c_r[y==categories[i]], x_d[y==categories[i]], space=space) for i in range(len(categories))])

	return max(h-wh, 0.0)




def least_mixed_conditional_mutual_information(x_c, y, z_c, x_d=None, z_d=None, space='dual', non_monotonic_extension=True):
	"""
	.. _least-mixed-conditional-mutual-information:
	Estimates the conditional mutual information between a dimensional random vector :math:`x` of inputs
	and a discrete scalar random variable :math:`y` (or label), conditional on a third random 
	variable :math:`z`, assuming the least amount of structure in :math:`(x, y, z)`, other than what is 
	necessary to be consistent with some observed properties.

	.. math::
		I(y; x_c, x_d | z_c, z_d) = I(y; x_c, x_d, z_c, z_d) - I(y; z_c, z_d)

		
	Parameters
	----------
	x_c : (n, d) np.array
		n i.i.d. draws from the generating distribution of continuous inputs.
	z_c : (n, d) np.array
		n i.i.d. draws from the generating distribution of continuous conditions.
	x_d : (n, d) np.array or None (default), optional
		n i.i.d. draws from the generating distribution of categorical inputs.
	z_d : (n, d) np.array or None (default), optional
		n i.i.d. draws from the generating distribution of categorical conditions.
	y : (n,) np.array
		n i.i.d. draws from the (categorical) labels generating distribution, sampled
		jointly with x.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.


	Returns
	-------
	i : float
		The mutual information between (x_c, x_d) and y, conditional on (z_c, z_d), **in nats**.

	Raises
	------
	AssertionError
		If y is not a one dimensional array or x and z are not all numeric.
	"""
	assert x_c is not None or z_c is not None, 'Both x_c and z_c cannot be None'
	assert x_c is not None or x_d is not None, 'Both x_c and x_d cannot be None'
	assert z_c is not None or z_d is not None, 'Both z_c and z_d cannot be None'
	if z_c is None and z_d is not None and x_c is not None:
		assert np.can_cast(x_c, float), 'x_c should be continuous'
		# I(y; x_c | z_d, x_d) = \sum_i p_i I(y; x_c | x_d, z_d = i)
		ds = z_d if x_d is None else np.hstack((x_d, z_d))
		ds = np.array(['*_*'.join(list(_)) for _ in ds])
		ds_cats, ds_counts = np.unique(ds)
		ds_probas = ds_counts/ds.shape[0]
		cmi = np.sum([ds_probas[i]*least_mixed_mutual_information(\
						x_c[ds==ds_cats[i]], y[ds==ds_cats[i]].flatten(), \
						x_d=None, space=space, non_monotonic_extension=non_monotonic_extension) \
						for i in range(ds_cats.shape[0])])

		return max(cmi, 0.0)


	if x_c is None and x_d is not None and z_c is not None:
		assert np.can_cast(z_c, float), 'z_c should be continuous'
		# I(y; x_d | z_c, z_d) = I(y; x_d, z_c, z_d) - I(y; z_c, z_d)
		ds = x_d if z_d is None else np.hstack((x_d, z_d))
		ds = np.array(['*_*'.join(list(_)) for _ in ds])
		cmi = least_mixed_mutual_information(\
						z_c, y.flatten(), \
						x_d=ds, space=space, \
						non_monotonic_extension=non_monotonic_extension)

		cmi -= least_mixed_mutual_information(\
						z_c, y.flatten(), \
						x_d=np.array(['*_*'.join(list(_)) for _ in x_d]), space=space, \
						non_monotonic_extension=non_monotonic_extension)

		return max(cmi, 0.0)

	# I(y; x_c, x_d | z_c, z_d) = I(y; x_c, x_d, z_c, z_d) - I(y; z_c, z_d)
	assert np.can_cast(x_c, float) and np.can_cast(z_c, float), 'z_c should be continuous'
	ds = None if x_d is None and z_d is None else x_d if z_d is None else z_d if x_d is None else np.hstack((x_d, z_d))
	ds = None if ds is None else np.array(['*_*'.join(list(_)) for _ in ds])

	x_c_ = x_c[:, None] if len(x_c.shape) == 1 else x_c
	z_c_ = z_c[:, None] if len(z_c.shape) == 1 else z_c
	cs = np.hstack((x_c_, z_c_))

	cmi = least_mixed_mutual_information(\
					cs, y.flatten(), \
					x_d=ds, space=space, \
					non_monotonic_extension=non_monotonic_extension)

	cmi -= least_mixed_mutual_information(\
					z_c, y.flatten(), \
					x_d=z_d, space=space, \
					non_monotonic_extension=non_monotonic_extension)

	return max(cmi, 0.0)




def discrete_mutual_information(x, y):
	"""
	.. _discrete-mutual-information:
	Estimates the (Shannon) mutual information between two discrete random variables from
	i.i.d. samples, using the plug-in estimator of Shannon entropy.

	Parameters
	----------
	x : (n,) np.array or None (default)
		n i.i.d. draws from a discrete distribution.
	y : (n,) np.array
		n i.i.d. draws from another discrete distribution, sampled jointly with x.

	Returns
	-------
	i : float
		The mutual information between x and y, **in nats**.

	Raises
	------
	AssertionError
		If y or x is not a one-dimensional array, or if x and y have different shapes.
	"""
	assert len(y.shape) == 1 or y.shape[1] == 1, 'Only one-dimensional outputs are supported for now.'
	assert len(x.shape) == 1 or x.shape[1] == 1, 'Only one-dimensional outputs are supported for now.'
	assert x.shape[0] == y.shape[0], 'Both arrays should have the same dimension.'

	hx = discrete_entropy(x)
	hy = discrete_entropy(y)
	hxy = discrete_entropy(np.array([str(x[i]) + '*_*' + str(y[i]) for i in range(x.shape[0])]))

	return max(0., hx+hy-hxy)


